# Copyright (C) 2013 Duncan Macleod
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Interface to the LIGO Channel Information System

Queries can be made using any part of a channel name, returning a list
of entries (key,value dicts) of found channels::

>>> from glue import cis
>>> out = cis.query('PSL-ISS_PSB_OUT_DQ')
>>> print([channel['name'] for channel in out])
[u'H1:PSL-ISS_PDB_OUT_DQ', u'H2:PSL-ISS_PDB_OUT_DQ', u'L1:PSL-ISS_PDB_OUT_DQ']

(example accessed August 26 2013).
"""

import json
import numpy
from six.moves import urllib

from glue.auth.saml import HTTPNegotiateAuthHandler

from . import (git_version, auth)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__version__ = git_version.id
__date__ = git_version.date

CIS_API_URL = 'https://cis.ligo.org/api/channel'
CIS_DATA_TYPE = {4: numpy.float32}


def query(name, debug=False):
    """Query the Channel Information System for details on the given
    channel name

    Example::

    >>> from glue import cis
    >>> out = cis.query('PSL-ISS_PSB_OUT_DQ')
    >>> print([channel['name'] for channel in out])
    [u'H1:PSL-ISS_PDB_OUT_DQ', u'H2:PSL-ISS_PDB_OUT_DQ',
     u'L1:PSL-ISS_PDB_OUT_DQ']

    Parameters
    ----------
    name : str
        Name of the channel of interest, or part of channel to query

    Returns
    -------
    list
        list of (key, value) dicts recovered from the CIS

    Raises
    ------
    ValueError
        if no channels are found matching the given name
    """
    url = '%s/?q=%s' % (CIS_API_URL, name)
    more = True
    out = []
    while more:
        reply = _get(url, debug=debug)
        if reply['count'] == 0:
            raise ValueError("No channels found with name '%s'" % name)
        try:
           out.extend(reply[u'results'])
        except KeyError:
           pass
        more = 'next' in reply and reply['next'] is not None
        if more:
            url = reply['next']
        else:
            break
    out.sort(key=lambda c: c['name'])
    return out


def _get(url, debug=False):
    """Perform a GET query against the CIS
    """
    try:
        response = auth.request_ligodotorg(url, debug=debug)
    except urllib.error.HTTPError:
        raise ValueError("Channel named '%s' not found in Channel "
                         "Information System. Please double check the "
                         "name and try again." % url.strip('=')[-1])
    return json.loads(response.read())



# Copyright (C) 2006--2009,2012--2016  Kipp Cannon
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
High-level support for Param elements.
"""


import re
import sys
from xml.sax.saxutils import escape as xmlescape
from xml.sax.xmlreader import AttributesImpl as Attributes
try:
	import yaml
except ImportError:
	# yaml serialization is optional
	pass


from glue import git_version
from . import ligolw
from . import types as ligolwtypes


__author__ = "Kipp Cannon <kipp.cannon@ligo.org>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


#
# =============================================================================
#
#                                  Utilities
#
# =============================================================================
#


def get_param(xmldoc, name):
	"""
	Scan xmldoc for a param named name.  Raises ValueError if not
	exactly 1 such param is found.
	"""
	params = Param.getParamsByName(xmldoc, name)
	if len(params) != 1:
		raise ValueError("document must contain exactly one %s param" % Param.ParamName(name))
	return params[0]


def get_pyvalue(xml, name):
	"""
	Convenience wrapper for get_param() that recovers an instance of a
	Python builtin type from a Param element.
	"""
	# Note:  the Param is automatically parsed into the correct Python
	# type, so this function is mostly a no-op.
	return get_param(xml, name).pcdata


#
# =============================================================================
#
#                               Element Classes
#
# =============================================================================
#


#
# FIXME: params of type string should be quoted in order to correctly
# delimit their extent.  If that were done, then the pcdata in a Param
# element could be parsed using the Stream tokenizer (i.e., as though it
# were a single-token stream), which would guarantee that Stream data and
# Param data is parsed using the exact same rules.  Unfortunately, common
# practice is to not quote Param string values, so we parse things
# differently here.  In particular, we strip whitespace from the start and
# stop of all Param pcdata.  If this causes your string Param values to be
# corrupted (because you need leading and trailing white space preserved),
# then you need to make everyone switch to quoting their string Param
# values, and once that is done then this code will be changed.  Perhaps a
# warning should be emitted for non-quoted strings to encourage a
# transition?
#


class Param(ligolw.Param):
	"""
	High-level Param element.  The value is stored in the pcdata
	attribute as the native Python type rather than as a string.
	"""
	class ParamName(ligolw.LLWNameAttr):
		dec_pattern = re.compile(r"(?P<Name>[a-z0-9_:]+):param\Z")
		enc_pattern = u"%s:param"

	Name = ligolw.attributeproxy(u"Name", enc = ParamName.enc, dec = ParamName)
	Scale = ligolw.attributeproxy(u"Scale", enc = ligolwtypes.FormatFunc[u"real_8"], dec = ligolwtypes.ToPyType[u"real_8"])
	Type = ligolw.attributeproxy(u"Type", default = u"lstring")

	def endElement(self):
		if self.pcdata is not None:
			# convert pcdata from string to native Python type
			if self.Type == u"yaml":
				try:
					yaml
				except NameError:
					raise NotImplementedError("yaml support not installed")
				self.pcdata = yaml.load(self.pcdata)
			else:
				self.pcdata = ligolwtypes.ToPyType[self.Type](self.pcdata.strip())

	def write(self, fileobj = sys.stdout, indent = u""):
		fileobj.write(self.start_tag(indent))
		for c in self.childNodes:
			if c.tagName not in self.validchildren:
				raise ligolw.ElementError("invalid child %s for %s" % (c.tagName, self.tagName))
			c.write(fileobj, indent + ligolw.Indent)
		if self.pcdata is not None:
			if self.Type == u"yaml":
				try:
					yaml
				except NameError:
					raise NotImplementedError("yaml support not installed")
				fileobj.write(xmlescape(yaml.dump(self.pcdata).strip()))
			else:
				# we have to strip quote characters from
				# string formats (see comment above).  if
				# the result is a zero-length string it
				# will get parsed as None when the document
				# is loaded, but on this code path we know
				# that .pcdata is not None, so as a hack
				# until something better comes along we
				# replace zero-length strings here with a
				# bit of whitespace.  whitespace is
				# stripped from strings during parsing so
				# this will turn .pcdata back into a
				# zero-length string.  NOTE:  if .pcdata is
				# None, then it will become a zero-length
				# string, which will be turned back into
				# None on parsing, so this mechanism is how
				# None is encoded (a zero-length Param is
				# None)
				fileobj.write(xmlescape(ligolwtypes.FormatFunc[self.Type](self.pcdata).strip(u"\"") or u" "))
		fileobj.write(self.end_tag(u"") + u"\n")

	@classmethod
	def build(cls, name, Type, value, start = None, scale = None, unit = None, dataunit = None, comment = None):
		"""
		Construct a LIGO Light Weight XML Param document subtree.
		FIXME: document keyword arguments.
		"""
		elem = cls()
		elem.Name = name
		elem.Type = Type
		elem.pcdata = value
		# FIXME:  I have no idea how most of the attributes should be
		# encoded, I don't even know what they're supposed to be.
		if dataunit is not None:
			elem.DataUnit = dataunit
		if scale is not None:
			elem.Scale = scale
		if start is not None:
			elem.Start = start
		if unit is not None:
			elem.Unit = unit
		if comment is not None:
			elem.appendChild(ligolw.Comment()).pcdata = comment
		return elem

	@classmethod
	def from_pyvalue(cls, name, value, **kwargs):
		"""
		Convenience wrapper for .build() that constructs a Param
		element from an instance of a Python builtin type.  See
		.build() for a description of the valid keyword arguments.
		"""
		if value is not None:
			return cls.build(name, ligolwtypes.FromPyType[type(value)], value, **kwargs)
		return cls.build(name, None, None, **kwargs)

	@classmethod
	def getParamsByName(cls, elem, name):
		"""
		Return a list of params with name name under elem.
		"""
		name = cls.ParamName(name)
		return elem.getElements(lambda e: (e.tagName == cls.tagName) and (e.Name == name))


#
# =============================================================================
#
#                               Content Handler
#
# =============================================================================
#


#
# Override portions of a ligolw.LIGOLWContentHandler class
#


def use_in(ContentHandler):
	"""
	Modify ContentHandler, a sub-class of
	glue.ligolw.LIGOLWContentHandler, to cause it to use the Param
	class defined in this module when parsing XML documents.

	Example:

	>>> from glue.ligolw import ligolw
	>>> def MyContentHandler(ligolw.LIGOLWContentHandler):
	...	pass
	...
	>>> use_in(MyContentHandler)
	"""
	def startParam(self, parent, attrs):
		return Param(attrs)

	ContentHandler.startParam = startParam

	return ContentHandler



# Copyright (C) 2010 Peter F. Couvares, Nickolas Fotopoulos
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Compare two LIGO LW XML files.
"""

from __future__ import print_function
import sys
import optparse
import os

from glue import git_version
from glue.ligolw import table
from glue.ligolw import utils

__author__ = "Peter Couvares <peter.couvares@ligo.org>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date

# returns True if table1 and table2 contents are identical; False otherwise
def diff_tables(table1, table2):
    result = True
    columns = table1.getElementsByTagName(table.Column.tagName)
    # TODO: correctly handle the case where table2 is a superset of table1
    for column1 in columns:
        cname = column1.getAttribute("Name")
        if cname in opts.exclude_columns.split(','):
            if opts.debug: print("skipping %s" % cname, file=sys.stderr)
            continue
        column2 = table2.getColumnByName(cname)
        if diff_columns(column1, column2) is False:
            # if we're reporting column diffs, note the false result and continue
            if opts.columns:
                result = False
            # otherwise we're only reporting table diffs, and we now
            # know the tables differ, so can stop comparing and return
            else:
                return False
    return result
                                                                                
# returns True if colA and colB contents are identical; False otherwise
def diff_columns(colA, colB):
    cname = colA.getAttribute("Name")
    if opts.debug: print("*** comparing %s ..." % cname)
    if len(colA) != len(colB):
        if opts.columns: print("%s differs" % cname)
        return False
    for i, row in enumerate(colA):
        if row != colB[i]:
            if opts.columns: print("%s differs" % cname)
            return False
    return True

# parse command line
usage = "Usage: %s doc1.xml doc2.xml" % sys.argv[0]
parser = optparse.OptionParser(usage=usage)
parser.add_option("--debug", action="store_true", default=False,
    help="Print debugging output")
parser.add_option("--tables", action="store_true", default=False,
    help="Print list of differing tables")
parser.add_option("--columns", action="store_true", default=False,
    help="Print list of differing columns within each table")
parser.add_option("--exclude-tables", metavar="CSV_LIST", default="",
    help="Ignore tables in CSV_LIST")
parser.add_option("--exclude-columns", metavar="CSV_LIST", default="",
    help="Ignore columns in CSV_LIST")

opts, args = parser.parse_args()
if len(args) != 2:
    print("Error: require exactly two arguments", file=sys.stderr)
    print(usage)
    sys.exit(2)

if opts.columns and opts.tables:
    print("Error: --columns and --tables are mutually exclusive; please choose one", file=sys.stderr)
    print(usage)
    sys.exit(2)

filename1, filename2 = args
if not os.path.exists(filename1):
    print("Error: %s does not exist" % filename, file=sys.stderr)
    sys.exit(2)
if not os.path.exists(filename2):
    print("Error: %s does not exist" % filename, file=sys.stderr)
    sys.exit(2)

# parse documents
xmldoc1 = utils.load_filename(filename1)
xmldoc2 = utils.load_filename(filename2)

exit_code = 0
tables = xmldoc1.getElementsByTagName(table.Table.tagName)
# TODO: correctly handle the case where xmldoc2 is a superset of xmldoc1
for t1 in tables:
    tname = t1.getAttribute("Name")

    if tname in opts.exclude_tables.split(','):
        if opts.debug: print("*** skipping %s ..." % tname, file=sys.stderr)
        continue

    if opts.debug: print("*** comparing %s ..." % tname)

    t2 = table.get_table(xmldoc2, tname)
    if diff_tables(t1, t2) is False:
        exit_code = 1
        if opts.tables:
            print("%s differs" % tname)
        elif opts.columns is False:
            # if we don't need to enumerate table or column diffs, we're done
            break

sys.exit(exit_code)



from __future__ import print_function
import os,sys
import matplotlib
matplotlib.use('Agg')
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
import numpy as np

#== Originally written by: Justin Garofoli justing@physics.syr.edu
#== 2009 October 27

#== Maintained by: Duncan Macleod duncan.macleod@astro.cf.ac.uk

#== this update:
#== takes IFO as 2nd input argument (comma separated list).

"""
 todo:
 1) add logging
 2) put all output in a directory
 3) get flags that are on and their segments
      within the window, and draw those  (right now
      we don't know if a flag is on right after
      the central time.
"""

#== default IFOs options
IFOs = ('H1', 'L1', 'V1')

def get_segments(gpstime):
  req = 'ligolw_dq_query -d --report -i ' + str(gpstime) 
  pp = os.popen(req)
  dd = {}
  for line in pp.readlines():
    if line[0:2] in IFOs:
      print(line[:-1])
    this_line = line[:-1]
    these_words = this_line.split()
    dd[these_words[0]] = (int(these_words[1][1:]), int(these_words[2]), int(these_words[3][:-1]))
  return dd

def get_segments_test(gpstime):
  pp = os.popen('cat test.txt')
  dd = {}
  for line in pp.readlines():
    this_line = line[:-1]
    these_words = this_line.split()
    dd[these_words[0]] = (int(these_words[1][1:]), int(these_words[2]), int(these_words[3][:-1]))
  return dd

def make_figure(dd,width):
  kk = list(dd.keys())
  kk.sort()
  for ifo in IFOs:
    fig = Figure()
    fig.set_figwidth(12.0)
    fig.set_figheight(6.0)
    ax = fig.add_axes([0.1, 0.2, 0.8, 0.7])
    i = 0
    for k in kk:
      if ifo == k[0:2]:
        times = dd[k]

        start = times[0] - times[1]
        if start < (-1*width):
          start = -1*width

        stop  = times[2] - times[1]
        if stop > width:
          stop = width

        stop = float(stop)
        start = float(start)
        t = np.linspace(start,stop,stop-start)
        y = i * np.ones(np.shape(t))
        cc = 'r'
        if k.find('UP:') > -1 or k.find('LIGHT:') > -1 or k.find('SCIENCE:') > -1 or k.find('CALIBRATED:') > -1:
          cc = 'g'
        elif k.find('V1:') > -1:
          cc = 'm'
        line, = ax.plot(t,y,label=k,c=cc,lw=2)
        ha = 'right'
        if (-1)**i==1:
          ha = 'left'
        print(str(i) + ' ' + k + ' ' + str(t.shape[0]) + ' ' + ha)
        ax.text(((-1)**i)*0.1*width,i,k,bbox=dict(facecolor='white', alpha=0.8, edgecolor='white'),horizontalalignment=ha, verticalalignment='center',fontsize=10)
        i = i+1
    ax.plot((0,0),(-1,i),c='k',lw=1,zorder=-1)
    ax.set_xlim(-1*width,width)
    ax.set_title(ifo + ' data quality flags at t0 = ' + str(times[1]))
    ax.set_ylim(-1,i)
    ax.set_yticks(np.arange(-1,i),''*(i+1))
    ax.set_xlabel('time after t0 (seconds)')
    canvas = FigureCanvasAgg(fig)
    canvas.print_figure(ifo + '-' + str(times[1]) + '_flags.png',dpi=150)
  return 1

if __name__ == '__main__':
  if len(sys.argv)==1:
    print('error, must give gpstime argument')
    sys.exit()
  elif len(sys.argv)==2:
    gpstime = sys.argv[1]
    IFOs = ['H1','L1','V1']
    width=600
  elif len(sys.argv)==3:
    gpstime = sys.argv[1]
    IFOs    = sys.argv[2].split(',')
    width=600
  elif len(sys.argv)==4:
    gpstime = sys.argv[1]
    IFOs    = sys.argv[2].split(',')
    width   = float(sys.argv[3])    
  else :
    print('wrong arguments')
    sys.exit()

  #print sys.argv[1]
  #print sys.argv
  dd = get_segments(gpstime)
  #dd = get_segments_test(931540290)
  make_figure(dd,width)




"""
This module defines the some useful functions for GSI socket servers.

This file is part of the Grid LSC User Environment (GLUE)

GLUE is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__author__ = "Scott Koranda <skoranda@gravity.phys.uwm.edu>"
from glue import git_version
__date__ = git_version.date
__version__ = git_version.id

import os
import sys
import re
from pyGlobus import security

class Gridmap(object):
  """
  Class to represent the grid-mapfile used for the server. 
  Methods __getitem__ and __setitem__ are added for convenience
  so that an instance can be treated like a dictionary.

  Entries in the grid-mapfile should be on a single line and
  surrounded by double quotes. No mapping to a local user ID 
  is necessary.
  """
  def __init__(self, filePath=None, logger=None):
    """
    Initialize a new instance. If filePath is None then check
    the environment for the variable GRIDMAP and uses the 
    path found as the path to the grid-mapfile to open and parse.
    
    @return: an instance that can be treated like a dictionary
    """
    if not filePath:
      try: 
        path = os.environ["GRIDMAP"]
      except:
        if logger:
          logger.error("Error parsing GRIDMAP from environment: %s" % e)
          logger.error("Authentication to server will fail for all subjects")
    else:   
      path = filePath
      self.path = path
      self.logger = logger


  def __getitem__(self, key):
    f = None
    try:
      f = open(self.path, 'r')
    except Exception as e:
      if self.logger:
        self.logger.error("Error opening grid-mapfile %s: %s" % (self.path, e))
        self.logger.error("Authentication to server will fail for all subjects")
                
    try:
      subjects = []
      lines = [ s.strip() for s in f.readlines()]
      for s in lines:
        a = s.split('"')
        for b in a:
          if b == '': continue
          else:
            subjects.append(b)
            break
    except Exception as e:
      if self.logger:
        self.logger.error("Error parsing grid-mapfile %s: %s" % (self.path, e))
        self.logger.error("Authentication to server may fail for some subjects")

    self.d = {} 
    for s in subjects:
      self.d[s] = 1

    if f:
      f.close()

    if key in self.d: return 1
    else: return 0


class AuthCallback(object):
  def __init__(self, gridmap, logger, callback=None):
    self.gridmap = gridmap
    self.logger = logger
    self.callback = callback

  def __call__(self, arg, handle, identity, context):
    """
    Callback function for GSI authentication. Each time this 
    function is called a new instance of the GridMap class is created
    that reads the grid-mapfile. If the subject passed in as the identity
    is in the grid-mapfile then 1 is returned and the connection is
    authenticated and allowed to continue. If the subject passed is
    not in grid-mapfile then a 0 is returned and the connection is
    broken.

    This function is supplied to help create an instance of io.AuthData
    which is then supplied to help create an instance of io.TCPIOAttr,
    which is then supplied to help create the instance of 
    io.ThreadingGSITCPSocketServer. See the
    U{pyGlobus<http://www-itg.lbl.gov/gtg/projects/pyGlobus/>} documentation.

    Everything is within a try/except clause since if an exception is raised
    within this function experience shows that the connection will be
    allowed to authenticate!

    @param handle: handle to the io.AuthData instance, not used here
    @param identity: the identity of the party trying to authenticate
    @type identity: string
    @param context: the GSI context, not used here

    @return: 1 to accept and 0 to reject
    """
    try:
      # load gridmap file
      if self.gridmap[identity]:
        if self.logger:
          self.logger.info("Accepted connection from %s" % identity)
        if self.callback:
          self.callback(arg, 1)
        return 1
      else:
        if self.logger:
          self.logger.info("Rejected connection from %s" % identity)
        if self.callback:
          self.callback(arg, 0)
        return 0
    except Exception as e:
      try:
        self.logger.error("Error within authentication callback: %s" % e)
      except Exception as e:
        pass
      return 0


def daemon():
  """
  Common code for making a process into a daemon. 

  Fork once, change directory to /, get a new session, set the umask to 0,
  then fork again.

  @return: Parent never returns. The child gets None.
  """
  # do the first fork
  try:
    pid = os.fork()
    if pid > 0:
      # exit first parent
      sys.exit(0)
  except OSError as e:
    sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
    sys.exit(1)
        
  # decouple from parent environment
  os.chdir("/")
  os.setsid()
  os.umask(0)
        
  # do the second fork
  try:
    pid = os.fork()
    if pid > 0:
      # exit from second parent
      sys.exit(0)
  except OSError as e:
    sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
    sys.exit(1)
        
  # send stdout and stderr to /dev/null
  try:
    devnull = open("/dev/null", "w")
    sys.stdout = devnull
    sys.stderr = devnull
  except Exception as e:
    sys.__stderr__.write("Unable to direct to /dev/null: %s\n" % e)
    sys.exit(1)


def checkCredentials():
    """
    Check to make sure that the proper Grid Credentials (a proxy certificate)
    is available in order to authenticate to the remote LSCsegFindServer.
    """
    # verify that we have access to credentials
    try:
      proxyText = security.grid_proxy_info()
    except Exception as e:
      sys.stderr.write("Error verifying credentials: %s\n" % e)
      sys.stderr.write("Run 'grid-proxy-init' to generate a proxy certificate\n")
      return False

    pat = re.compile(r'timeleft : (\d{1,3}):(\d\d):(\d\d)')

    try:
      if isinstance(proxyText, str):
        m = pat.search(proxyText)
      elif isinstance(proxyText, tuple):
        m = pat.search(proxyText[0])
      else:
        raise RuntimeError("bad format for proxyText in checkCredentials")
      hours, minutes, seconds = map(int, m.groups())

    except Exception as e:
      sys.stderr.write("Error parsing proxy information: %s\n" % e)
      return False

    timeleft = seconds + 60 * minutes + 3600 * hours

    if timeleft < 300:
      sys.stderr.write("Less than 5 minutes left for proxy certificate.\n")
      sys.stderr.write("Run 'grid-proxy-init' to generate a new proxy certificate\n")
      return False

    return True




from __future__ import print_function
import subprocess
import shutil
import os, sys
import copy
from optparse import OptionParser

from ligo import segments
from glue.ligolw import ligolw
from glue.ligolw import utils
from glue.ligolw import table
from glue.ligolw import lsctables
from glue.ligolw.utils import ligolw_add
from glue.ligolw.utils import process
from glue.ligolw.utils import print_tables
from glue import git_version
from glue.segmentdb import query_engine
from glue.segmentdb import segmentdb_utils

from pylal.printutils import format_end_time_in_utc
from pylal import InspiralUtils

__prog__ = "ligolw_veto_def_check"
__author__ = "Collin Capano <cdcapano@physics.syr.edu>"

#
#   Parse command line
#

parser = OptionParser(usage = "%s [options] veto_file1.xml veto_file2.xml ... " % __prog__, description = "Checks the segment database for higher versions of flags in the given veto definer file(s) during the specified times. Outputs a summary of what flags have higher versions and for what segments. Also checks if SciMon flags in the veto-definer file(s) are in the database. If the optional --write-new-veto-file is specified, a new veto definer file will be written based on what the program thinks are the correct versions to use. That is: vetoes that are found to have higher versions in the database will be re-written with the higher version. If a veto is found to not intersect at all with the correct version in the segment database, it will be discarded. This action can be overridden for specific flags using the various skip- options, however.")

parser.add_option("-o", "--output-directory", default = "./", help = "write summary files to directory. Default is to write to the current directory.")
parser.add_option("-n", "--write-new-veto-file", action = "store_true", default = False,  help = "write a new veto definer file with the updated versions to the specified files. The file will be named {IFOS}-UPDATED_VETO_DEFINER-{START_TIME}-{DURATION}.xml, where IFOS are all the ifos in the input veto definer file(s), START_TIME is the start time of the analysis (see --start-time for more info), and DURATION is the duration of the analysis (see --end-time for more info). This file will be located in the output-directory.")
parser.add_option("-s", "--start-time", action = "store", type = "int", default = 931035296, help = "earliest time to check; default is the start of S6 (931035296)")
parser.add_option("-e", "--end-time", action = "store", type = "int", default = None, help = "latest time to check; default is to check up to now. WARNING: If set to anything other than the default (now), the analysis may not be correct. This is because higher versions of flags may be defined at a point later than the given end time, making any lower version found obsolete.")
parser.add_option("-F", "--skip-flag", action = "append", default = [], metavar = "IFO:NAME", help = "Keep all of the original vetoes for a given flag the same in the updated veto-definer file, regardless of the results of the analysis. Note that the flag will still be analyzed, so this option only has an effect if a new veto-definer file is desired. To specify multiple flags, give the option multiple times." )
parser.add_option("-I", "--skip-ifo", action = "append", default = [], metavar = "IFO", help = "Add all of the given ifo's flags to the list of skipped flags. To specify multiple flags, give the option multiple times." )
parser.add_option("-T", "--skip-type", action = "append", default = [], metavar = "TYPE", help = "Add all flags of the given type to the list of skipped flags. A flag's type is given by whatever preceeds the first '-' in the flag name. For example, H1:DMT-UP is a DMT flag. To specify multiple flags, give the option multiple times. Note: this does not work for Virgo flags." )
parser.add_option("-S", "--skip-science-linked-flags", action = "store_true", default = False, help = "Add Science-related DMT flags to the list of skipped flags. These are flags that are republished along with DMT-SCIENCE, e.g., DMT-BADGAMMA." )
parser.add_option("-U", "--skip-undefined-flags", action = "store_true", default = False, help = "Add undefined flags to the list of skipped flags. These are non-SciMon flags that are found to not intersect any of the vetoes in the input file(s). This can be due to a mistake in either the veto-definer file or in the insertion of the flag into the segment database. The default action is to delete these from the updated veto-definer file; turning this on will cause the vetoes to be left in as they were in the original file(s)." )
parser.add_option("-v", "--verbose", action = "store_true", default = False, help = "Be verbose.")

opts, veto_def_files = parser.parse_args()
# check options
veto_def_files = [x != 'stdin' and x or None for x in veto_def_files]
if veto_def_files == []:
    veto_def_files = [None]
if opts.end_time is not None:
    print("WARNING: By setting an end-time to analyze, this analysis may not be correct. If a higher version of any flag has been inserted after the given end time, those flags will be obsolete.", file=sys.stderr)
# adjust list arguments to comma separated strings so as not to confuse process_params table
skip_flags = []
skip_ifos = []
skip_types = []
for listopt, listname in [('skip_ifo', skip_ifos), ('skip_type', skip_types), ('skip_flag',skip_flags)]:
    if getattr(opts, listopt) != []:
        listname.extend( list(map(str.upper, getattr(opts, listopt))) )
        setattr(opts, listopt, ','.join([flag for flag in listname]))
    else:
        setattr(opts, listopt, None)
# make skip_flags into a list of tuples
skip_flags = [tuple(x.split(':')) for x in skip_flags]
    

#
#   Main
#

# copy each input file to the output directory and get md5sums of each
analyzed_files = {}
for filename in veto_def_files:
    # if input is stdin (i.e., filenmae is None), copy the input to a file called 'stdin_input.xml' to the output directory
    if filename is None:
        target_file = '%s/stdin_input.xml' % opts.output_directory 
        f = open( target_file, 'w')
        shutil.copyfileobj( sys.stdin, f )
        f.seek(0)
    else:
        target_file = "%s/%s" % (opts.output_directory, os.path.basename(filename))
        if not shutil._samefile(filename, target_file):
            shutil.copyfile( filename, target_file )
        f = open(target_file, 'r')
    # get the md5sum
    _, md5sum = utils.load_fileobj( f )
    f.close()
    # add to the files to analyze
    analyzed_files[target_file] = md5sum

# load the veto definer file(s)
xmldoc = ligolw_add.ligolw_add( ligolw.Document(), list(analyzed_files.keys()) )
veto_def_table = lsctables.VetoDefTable.get_table(xmldoc)

# create summary_file directory in output directory to store files to
if not os.path.exists( '%s/%s' %( opts.output_directory, 'summary_files' ) ):
    os.mkdir( '%s/%s' %( opts.output_directory, 'summary_files') )

# connect to the segment database
database_location = "https://segdb.ligo.caltech.edu"
connection = segmentdb_utils.setup_database(database_location)
engine     = query_engine.LdbdQueryEngine(connection)

# create a veto definer table to write to
outtable = lsctables.New(lsctables.VetoDefTable)
outdoc = ligolw.Document()
outdoc.appendChild(ligolw.LIGO_LW())
outproc_id = process.register_to_xmldoc(outdoc, __prog__, opts.__dict__, version = git_version.id)

# freeze the time up to which we search; default is to set this to whatever the start time of the program is
if opts.end_time is not None:
    global_end_time = opts.end_time
else:
    global_end_time = outproc_id.start_time
global_segment = segments.segment(opts.start_time, global_end_time)

# add the original file's process info
origproctab = lsctables.ProcessTable.get_table(xmldoc)
newproctab = lsctables.ProcessTable.get_table(outdoc)
for proc in origproctab:
    newproc = lsctables.Process()
    for x in newproc.__slots__:
        try:
            setattr(newproc, x, getattr(proc,x))
        except AttributeError:
            # can happen if the process table in the original file is out of date
            setattr(newproc, x, None)
    newproctab.append(newproc)

# get html elements
nl = "<br />"
ttx, xtt, tx, xt, capx, xcap, rx, xr, cx, xc, rspx, xrsp, hlx, hxl, xhl = print_tables.set_output_format( "html" )
xt = "</table>"
thx = "<th>"
xth = "</th>"
ulx, xul, lix, xli = "<ul>", "</ul", "<li>", "</li>"

# check sci-mon flags; for these, just check that there is an intersecting segment in the segment database for the given name
found_scimon_flags = {}
missing_scimon_flags = {}
found_summ_files = {}
missing_summ_files = {}
if opts.verbose:
    print("Checking SciMon flags:", file=sys.stderr)
scimon_flags = dict([ [(veto.ifo, veto.name), [y for y in veto_def_table if y.ifo == veto.ifo and y.name == veto.name]] for veto in veto_def_table if veto.name.startswith('SCI-') ])
for this_ifo, this_name in sorted(scimon_flags):
    if opts.verbose:
        print("\t%s:%s" %(this_ifo, this_name), file=sys.stderr)
    for this_veto in scimon_flags[(this_ifo, this_name)]:
        vseg = segments.segment(this_veto.start_time, this_veto.end_time != 0 and this_veto.end_time or segments.infinity)
        sqlquery = "SELECT segdef.comment, segdef.version, segsumm.start_time, segsumm.end_time FROM segment_summary AS segsumm JOIN segment_definer AS segdef ON (segdef.segment_def_id = segsumm.segment_def_id) WHERE segdef.ifos = '%s' AND segdef.name = '%s' AND segsumm.end_time > %i" % (this_ifo, this_name, this_veto.start_time)
        for cmt, vn, start, end in engine.query(sqlquery):
            this_seg = segments.segment(start,end)
            if this_seg.intersects(vseg):
                found_scimon_flags.setdefault((this_ifo, this_name), {})
                found_scimon_flags[(this_ifo,this_name)].setdefault(this_veto, [])
                found_scimon_flags[(this_ifo, this_name)][this_veto].append((cmt, vn, this_seg))
        # check if this_veto has no counterpart in the segdb
        if (this_ifo, this_name) not in found_scimon_flags or ((this_ifo, this_name) in found_scimon_flags and this_veto not in found_scimon_flags[(this_ifo, this_name)]):
            missing_scimon_flags.setdefault((this_ifo,this_name), [])
            missing_scimon_flags[(this_ifo,this_name)].append(this_veto)

    # write found summary file
    if (this_ifo, this_name) in found_scimon_flags:
        summ_fn = "%s/summary_files/%s-%s_FOUND-%i-%i.html" % ( opts.output_directory, this_ifo, this_name, global_segment[0], abs(global_segment) )
        summ_file = open(summ_fn, 'w')
        print("%s\n%s" %( "<html>", "<body>" ), file=summ_file)
        print("<h1>%s:%s Flags Found in the SegDB</h1>" %( this_ifo, this_name), file=summ_file)
        print("<p>These are instances of %s in the checked file for which a corresponding flag is found in the segment database that intersects with it. All intersecting %s flags in the database are shown.</p>" %( ':'.join([this_ifo, this_name]), ':'.join([this_ifo, this_name]) ), file=summ_file)
        print('<table border="1", cellpadding="5">', file=summ_file)
        print(rx + '<th colspan="7">From Veto-definer File</th><th colspan="5">From SegDB</th>' + xr, file=summ_file)
        print(rx + thx + (xth+thx).join([ 'Version', 'Category', 'Start Pad', 'End Pad', 'Start Time', 'End Time', 'Comment', 'Version', 'Start Time', 'End Time', 'Comment', 'Overlap' ]) + xth + xr, file=summ_file)
        for v in found_scimon_flags[(this_ifo, this_name)]:
            special_cx = "%s%i%s" % (rspx, len(found_scimon_flags[(this_ifo,this_name)][v]), xrsp) 
            print(rx + special_cx + (xc+special_cx).join([ str(v.version), str(v.category), str(v.start_pad), str(v.end_pad), str(v.start_time), str(v.end_time), v.comment ]) + xc, end=' ', file=summ_file)
            for n, (dbcmt, dbvn, dbseg) in enumerate(found_scimon_flags[(this_ifo, this_name)][v]):
                if n != 0:
                    print(rx, end=' ', file=summ_file)
                print(cx + (xc+cx).join([ str(dbvn), str(dbseg[0]), str(dbseg[1]), dbcmt, str(abs(dbseg & segments.segment(v.start_time, v.end_time == 0 and segment.infinity or v.end_time))) ]) + xc + xr, file=summ_file)
        print("</table>", file=summ_file)
        print("%s\n%s" %( "</body>", "</html>" ), file=summ_file)
        summ_file.close()
        found_summ_files[(this_ifo, this_name)] = summ_fn

    # write missing summary file
    if (this_ifo, this_name) in missing_scimon_flags:
        summ_fn = "%s/summary_files/%s-%s_MISSING-%i-%i.html" % ( opts.output_directory, this_ifo, this_name, global_segment[0], abs(global_segment) )
        summ_file = open(summ_fn, 'w')
        print("%s\n%s" %( "<html>", "<body>" ), file=summ_file)
        print("<h1>%s:%s Flags Not Found in the SegDB</h1>" %( this_ifo, this_name), file=summ_file)
        print("<p>These are all the %s entries in the veto definer file for which a corresponding flag could not be found in the segment database during the entries' specified time.</p>" %( ':'.join([this_ifo, this_name]) ) + nl, file=summ_file)
        print('<table style="background-color:#e5eecc", cellpadding="5">', file=summ_file)
        print(rx + thx + (xth+thx).join(["ifo", "name", "version", "category", "start_time", "end_time", "start_pad", "end_pad", "comment"]) + xth + xr, file=summ_file)
        for veto in missing_scimon_flags[(this_ifo, this_name)]:
            print(rx + cx + (xc+cx).join([veto.ifo, veto.name, str(veto.version), str(veto.category), str(veto.start_time), str(veto.end_time), str(veto.start_pad), str(veto.end_pad), veto.comment]) + xc + xr, file=summ_file)
        print("</table>", file=summ_file)
        print("%s\n%s" %( "</body>", "</html>" ), file=summ_file)
        summ_file.close()
        missing_summ_files[(this_ifo, this_name)] = summ_fn
        
# cycle over flags, analyzing
if opts.verbose:
    print("Checking the rest:", file=sys.stderr)

updated_flags = {}
missing_flags = []
unchanged_flags = []
undefined_flags = {}
non_intersecting_flags = {}

check_names = dict([ [(veto.ifo, veto.name), [y for y in veto_def_table if y.ifo == veto.ifo and y.name == veto.name]] for veto in veto_def_table if (veto.ifo,veto.name) not in scimon_flags])

# get flags to not update
other_skipped_flags = [(this_ifo, this_name) for this_ifo, this_name in check_names if this_ifo in skip_ifos or this_ifo != 'V1' and this_name.split('-')[0] in skip_types]
skip_flags.extend(other_skipped_flags)    

# FIXME: Virgo science flags?
science_related_flags = ['H1:DMT-INJECTION', 'H1:DMT-UP', 'H1:DMT-CALIBRATED', 'H1:DMT-BADGAMMA', 'H1:DMT-LIGHT', 'L1:DMT-SCIENCE', 'L1:DMT-INJECTION', 'L1:DMT-UP', 'L1:DMT-CALIBRATED', 'L1:DMT-BADGAMMA', 'L1:DMT-LIGHT']
if opts.skip_science_linked_flags:
    skip_flags.extend(tuple(x.split(':')) for x in science_related_flags)

for this_ifo, this_name in sorted(check_names):
    if opts.verbose:
        print("\t%s:%s" % (this_ifo, this_name), file=sys.stderr)
    # check if there are multiple versions
    sqlquery = "SELECT DISTINCT(version) FROM segment_definer JOIN segment_summary AS segsumm ON (segment_definer.segment_def_id = segsumm.segment_def_id) WHERE segment_definer.ifos = '%s' AND segment_definer.name = '%s' AND segsumm.end_time > %i" %( this_ifo, this_name, opts.start_time )
    versions = [row[0] for row in engine.query(sqlquery)]
    if len(versions) == 0:
        missing_flags.append((this_ifo, this_name))
    else:
        sqlquery = "SELECT segdef.version, segsumm.start_time, segsumm.end_time FROM segment_summary AS segsumm JOIN segment_definer AS segdef ON (segdef.segment_def_id = segsumm.segment_def_id) WHERE segdef.ifos = '%s' AND segdef.name = '%s' AND segsumm.end_time > %i" % (this_ifo, this_name, opts.start_time)
        segs = segments.segmentlistdict()
        for vn, start, end  in engine.query(sqlquery):
            seg = segments.segment(start,end)
            if seg.intersects(global_segment):
                segs.setdefault(vn, segments.segmentlist([]))
                segs[vn].append( segments.segment(start, end) & global_segment )
        
        # coalesce the segment lists, as well as create a copy to be used later to keep track of segments added by higher versions
        added_segs = segments.segmentlistdict([ [vn, copy.copy(sl.coalesce())] for vn, sl in segs.items() ])
        # check if the max version in the segment database is >= what's specified in the vetoes
        vdf_vn_too_high = max(segs.keys()) < max([veto.version for veto in check_names[(this_ifo, this_name)] ])
        if vdf_vn_too_high:
            for vn in [veto.version for veto in check_names[(this-ifo, this_name)] if veto.version > max(segs.keys())]:
                segs[vn] = segments.segmentlist([])
        # check if any of the segments intersect with the version given in the vdf
        flag_intersects = any([ segments.segment(veto.start_time, veto.end_time == 0 and segments.infinity or veto.end_time) in segs[veto.version] for veto in check_names[(this_ifo, this_name)] ])
        # now cycle over the versions going from highest to lowest and remove segments of higher versions to get the corrected segments
        versions = sorted([vn for vn in segs], reverse = True)
        seg_extents = {}
        uncertain_segs = segments.segmentlistdict()
        for vn in versions:
            seg_extents[vn] = segs[vn].extent()
            higher_versions = [higher_vn for higher_vn in versions if higher_vn > vn]
            for higher_vn in higher_versions:
                segs[vn] -= segs[higher_vn]
            # remove uncertain segs; these are segs that occur after all higher version's end times
            if higher_versions != []:
                uncertain_segs[vn] = segments.segmentlist([seg for seg in segs[vn] if all(seg_extents[higher_vn][1] <= seg[0] for higher_vn in higher_versions)])
                segs[vn] -= uncertain_segs[vn]

        # create the added segments segment list; this is done by going from low-to-high in versions and removing all segments from lower versions
        for vn in sorted(versions):
            for lower_vn in [x for x in sorted(versions) if x < vn]:
                added_segs[vn] -= added_segs[lower_vn]
        
        # create updated veto entries
        unique_vetoes = set([(this_ifo, this_name, veto.category, veto.start_pad, veto.end_pad, veto.comment, veto.start_time, veto.end_time) for veto in check_names[(this_ifo, this_name)]])
        added_flags = []
        for ifo, name, cat, start_pad, end_pad, comment, orig_start_time, orig_end_time in unique_vetoes:
            veto_seg = segments.segmentlist([segments.segment(orig_start_time, orig_end_time == 0 and segments.infinity or orig_end_time)])
            new_flag_segs = {}
            # create flag from the highest version
            if veto_seg.intersects( segments.segmentlist([seg_extents[max(versions)]]) ):
                new_flag_segs[max(versions)] = (segments.segmentlist([seg_extents[max(versions)]]) & veto_seg).extent()
            # for lower versions, only add the uncertain segments
            for vn, sl in uncertain_segs.items():
                if veto_seg.intersects(sl):
                    new_flag_segs[vn] = (veto_seg & sl).extent()
            for vn, (new_start_time, new_end_time) in sorted(list(new_flag_segs.items()), reverse = True):
                updated_flag = lsctables.VetoDef()
                updated_flag.process_id = outproc_id.process_id
                updated_flag.ifo = ifo
                updated_flag.name = name
                updated_flag.version = vn 
                updated_flag.category = cat
                updated_flag.start_pad = start_pad
                updated_flag.end_pad = end_pad
                updated_flag.comment = comment
                if new_start_time != orig_start_time:
                    new_start_time -= abs(start_pad)
                updated_flag.start_time = new_start_time
                if orig_end_time == 0 and new_end_time == max([x[1] for x in new_flag_segs.values()]):
                    updated_flag.end_time = 0
                elif new_end_time != orig_end_time:
                    new_end_time += abs(end_pad)
                    updated_flag.end_time = new_end_time
                else:
                    updated_flag.end_time = new_end_time
                added_flags.append(updated_flag)
            
        # check if the flag in the skipped flags; if not, add the new ones to the outtable; otherwise, add the original
        if (this_ifo, this_name) in skip_flags or added_flags == [] and opts.skip_undefined_flags:
            for veto in check_names[this_ifo, this_name]:
                outtable.append(veto)
        else:
            for veto in added_flags:
                outtable.append(veto)

        # check how many vetoes have been updated
        def match_criteria(v):
            return (v.ifo, v.name, v.version, v.category, v.start_pad, v.end_pad, v.comment, v.start_time, v.end_time)
        # if all the original flags are the same as the new flags, add to unchanged_flags, otherwise, write a summary html file for it
        if all([ any([match_criteria(orig_veto) == match_criteria(new_veto) for new_veto in added_flags]) for orig_veto in check_names[(this_ifo,this_name)] ]):
            unchanged_flags.append((this_ifo,this_name))
        else:
            # write summary
            summ_fn = "%s/summary_files/%s-%s-%i-%i.html" % ( opts.output_directory, this_ifo, this_name, global_segment[0], abs(global_segment) )
            summ_file = open( summ_fn, 'w' )
            print("<html>\n<body>", file=summ_file)
            print("<h1> %s:%s </h1>" %( this_ifo, this_name ), file=summ_file)
            print("<p><i>This analysis valid for %i - %i (%s UTC - %s UTC)</i></p>" %( global_segment[0], global_segment[1], format_end_time_in_utc(global_segment[0]), format_end_time_in_utc(global_segment[1]) ), file=summ_file)
            print("<p>Used in vetoes:" + nl, file=summ_file)
            print('<table style="background-color:#e5eecc", cellpadding="5">', file=summ_file)
            print(rx + thx + (xth+thx).join(["ifo", "name", "version", "category", "start_time", "end_time", "start_pad", "end_pad", "comment"]) + xth + xr, file=summ_file)
            for veto in check_names[(this_ifo, this_name)]:
                print(rx + cx + (xc+cx).join([veto.ifo, veto.name, str(veto.version), str(veto.category), str(veto.start_time), str(veto.end_time), str(veto.start_pad), str(veto.end_pad), veto.comment]) + xc + xr, file=summ_file)
            print("</table>" + nl, file=summ_file)
           
            # check if these vetoes are undefined or not
            if added_flags == []:
                print('This flag is found in the segment database but none of the vetoes intersect with the times for which it was defined (i.e., segment_summary table segments). This is because:', file=summ_file)
                print('<table  style="background-color:#ff8080", celpadding="5"><tr><td><b>', file=summ_file) 
                if vdf_vn_too_high:
                    print('The minimum version specified in the veto-definer file is higher than the maximum version found in the segment database.', file=summ_file)
                elif flag_intersects:
                    print('The veto intersects with segment_summary segments in the segment database, but not with the highest version. This can result in time being vetoed that should not be (assuming the versions were republished correctly in the segment database).', file=summ_file)
                else:
                    print('The veto does not intersect with any versions of the flag in the segment database.', file=summ_file)
                print('</b></td></tr></table><br />', file=summ_file)
            else:
                print("These should be updated to:" + nl, file=summ_file)
                print('<table style="background-color:#90ee90", cellpadding="5">', file=summ_file)
                print(rx + thx + (xth+thx).join(["ifo", "name", "version", "category", "start_time", "end_time", "start_pad", "end_pad", "comment"]) + xth + xr, file=summ_file)
                for veto in added_flags:
                    print(rx + cx + (xc+cx).join([veto.ifo, veto.name, str(veto.version), str(veto.category), str(veto.start_time), str(veto.end_time), str(veto.start_pad), str(veto.end_pad), veto.comment]) + xc + xr, file=summ_file)
                print("</table>" + nl, file=summ_file)

            for vn in sorted(versions):
                start, end = seg_extents[vn]
                print("<b>Version %i extent:</b> %i - %i (%s - %s)" % ( vn, start, end, format_end_time_in_utc(start), format_end_time_in_utc(end) ) + nl, file=summ_file)
            print("</p>", file=summ_file)

            print("<h3>Extent of Uncertain Segments</h3>", file=summ_file)
            for vn, sl in sorted(uncertain_segs.items()):
                print(lix + "Version %i: " % vn, end=' ', file=summ_file)
                if uncertain_segs[vn] == []:
                    print("No uncertain segments." + xli, file=summ_file)
                else:
                    print("%i - %i (%s - %s)" %( sl.extent()[0], sl.extent()[1], format_end_time_in_utc(sl.extent()[0]), format_end_time_in_utc(sl.extent()[1]) ) + xli, file=summ_file)
            print(xul, file=summ_file)
            print('<p>This is the extent of segment_summary segments that come after the last defined segment for all higher versions than the stated version. These are "uncertain" because if new segments are appened to the higher versions in the future, these segments will be obsolete. For the individual segments, see "Uncertain Segment-summary Table Segments", below.</p>', file=summ_file)

            print("<hr /><h2>Summary of Deleted Segment-summary Table Segments</h2>", file=summ_file)
            print('<p>This is a summary of segment_summary table segments that were in the given version but not in higher versions. For the individual segments, see "Segment-summary Table Segments Deleted by Higher Versions", below.</p>', file=summ_file)
            for vn in sorted([x for x in versions if x != max(versions)]):
                print("<h3> Version %i </h3>" % vn, file=summ_file)
                print("<p><b>Total number of seconds deleted: %i</b>" % sum([abs(seg) for seg in segs[vn]]) + nl + nl, file=summ_file)
                print("Number of seconds deleted intersecting with above flag(s) in the veto definer file:", file=summ_file)
                print(tx, file=summ_file)
                print(''.join([ rx, thx, (xth+thx).join([ 'Category', 'Flag Start', 'Flag End', '# deleted seconds' ]), xth, xr ]), file=summ_file)
                for veto in [x for x in check_names[(this_ifo, this_name)] if x.version == vn]:
                    veto_seg = segments.segmentlist([segments.segment(veto.start_time, veto.end_time == 0 and segments.infinity or veto.end_time)])
                    print(''.join([ rx, cx, (xc+cx).join([ str(veto.category), str(veto.start_time), str(veto.end_time), str(veto_seg.intersects(segs[vn]) and sum([abs(seg) for seg in veto_seg & segs[vn]]) or 0) ]), xc, xr]), file=summ_file)
                print(xt + nl, file=summ_file)
                print("Number of seconds deleted prior to start of higher versions:", file=summ_file)
                print(tx, file=summ_file)
                print(''.join([ rx, thx, (xth+thx).join([ 'Higher Version #', 'Earliest start time', 'UTC', 'Number of seconds deleted prior' ]), xth, xr ]), file=summ_file)
                for higher_vn in sorted([x for x in versions if x > vn], reverse = True):
                    print(''.join([ rx, cx, (xc+cx).join([ str(higher_vn), str(seg_extents[higher_vn][0]), format_end_time_in_utc(seg_extents[higher_vn][0]), str(sum([abs(seg) for seg in segs[vn] if seg[1] <= seg_extents[higher_vn][0]])) ]), xc, xr ]), file=summ_file)
                print("%s\n%s" %(xt, nl), file=summ_file)

                print("For deleted segments in the extent of higher versions:" + ulx, file=summ_file)
                higher_vns_extent = segments.segmentlist([seg for higher_vn, seg in seg_extents.items() if higher_vn > vn]).coalesce().extent()
                print("%s number <= 2s long: %i %s" % (lix, len([seg for seg in segs[vn] if seg in higher_vns_extent and abs(seg) <= 2]), xli), file=summ_file)
                print("%s number > 2s and <= 64s long: %i %s" % (lix, len([seg for seg in segs[vn] if seg in higher_vns_extent and abs(seg) > 2 and abs(seg) <= 64]), xli), file=summ_file)
                print("%s number > 64s long: %i %s" % (lix, len([seg for seg in segs[vn] if seg in higher_vns_extent and abs(seg) > 64]), xli ), file=summ_file)
                print(xul + "</p>", file=summ_file)

            print("<hr /><h2>Summary of Added Segment-summary Table Segments</h2>", file=summ_file)
            print('<p>This is a summary of segment_summary table segments that were in the given version but not in lower versions. For individual segments, see "Segment-summary Table Segments Added by Lower Versions", below.</p>', file=summ_file)
            for vn in sorted([x for x in versions if x != min(versions)]):
                print("<h3> Version %i </h3>" % vn, file=summ_file)
                print("<p><b>Total number of seconds added: %i</b>" % sum([abs(seg) for seg in added_segs[vn]]) + nl + nl, file=summ_file)
                print("Number of seconds added intersecting with above flag(s) in the veto definer file:", file=summ_file)
                print(tx, file=summ_file)
                print(''.join([ rx, thx, (xth+thx).join([ 'Category', 'Flag Start', 'Flag End', '# added seconds' ]), xth, xr ]), file=summ_file)
                for veto in [x for x in check_names[(this_ifo, this_name)] if x.version == vn]:
                    veto_seg = segments.segmentlist([segments.segment(veto.start_time, veto.end_time == 0 and segments.infinity or veto.end_time)])
                    print(''.join([ rx, cx, (xc+cx).join([ str(veto.category), str(veto.start_time), str(veto.end_time), str(veto_seg.intersects(added_segs[vn]) and sum([abs(seg) for seg in veto_seg & added_segs[vn]]) or 0) ]), xc, xr]), file=summ_file)
                print(xt + nl, file=summ_file)

                print("For added segments in the extent of lower versions:" + ulx, file=summ_file)
                lower_vns_extent = segments.segmentlist([seg for lower_vn, seg in seg_extents.items() if lower_vn < vn]).coalesce().extent()
                print("%s number <= 2s long: %i %s" % (lix, len([seg for seg in added_segs[vn] if seg in lower_vns_extent and abs(seg) <= 2]), xli), file=summ_file)
                print("%s number > 2s and <= 64s long: %i %s" % (lix, len([seg for seg in added_segs[vn] if seg in lower_vns_extent and abs(seg) > 2 and abs(seg) <= 64]), xli), file=summ_file)
                print("%s number > 64s long: %i %s" % (lix, len([seg for seg in added_segs[vn] if seg in lower_vns_extent and abs(seg) > 64]), xli ), file=summ_file)
                print(xul + "</p>", file=summ_file)

            print("<hr /><h2> Uncertain Segment-summary Table Segments </h2>", file=summ_file)
            for vn, sl in sorted(uncertain_segs.items()):
                print("<h3> Version %i </h3>" % vn, file=summ_file)
                print(tx, file=summ_file)
                print("%s Segment %s UTC %s dt (s) %s" % ( rx+thx, xth+thx, xth+thx, xth+xr ), file=summ_file)
                for seg in sorted(uncertain_segs[vn]): #, key = lambda x: abs(x), reverse = True):
                    print(''.join([ rx, cx, "%i<br />%i" %( seg[0], seg[1]), xc, cx, format_end_time_in_utc(seg[0]), " <br /> ", format_end_time_in_utc(seg[1]), xc, cx, "%i" % abs(seg), xc, xr ]), file=summ_file)
                print(xt, file=summ_file) 

            print("<hr /><h2>Segment-summary Table Segments Deleted by Higher Versions </h2>", file=summ_file)
            print('<p>These are segment_summary table segments that were in the given version but not in higher versions.</p>', file=summ_file)
            for vn in sorted([x for x in versions if x != max(versions)]):
                print("<h3> Version %i </h3>" % vn, file=summ_file)
                print(tx, file=summ_file)
                print("%s Segment %s UTC %s dt (s) %s" % ( rx+thx, xth+thx, xth+thx, xth+xr ), file=summ_file)
                for seg in sorted(segs[vn]): #, key = lambda x: abs(x), reverse = True):
                    print(''.join([ rx, cx, "%i<br /> %i" %( seg[0], seg[1]), xc, cx, format_end_time_in_utc(seg[0]), "<br />", format_end_time_in_utc(seg[1]), xc, cx, "%i" % abs(seg), xc, xr ]), file=summ_file)
                print(xt, file=summ_file) 

            print("<hr /><h2>Segment-summary Table Segments Added by Lower Versions </h2>", file=summ_file)
            print('<p>These are segment_summary table segments that were in the given version but not in lower versions.</p>', file=summ_file)
            for vn in sorted([x for x in versions if x != min(versions)]):
                print("<h3> Version %i </h3>" % vn, file=summ_file)
                print(tx, file=summ_file)
                print("%s Segment %s UTC %s dt (s) %s" % ( rx+thx, xth+thx, xth+thx, xth+xr ), file=summ_file)
                for seg in sorted(added_segs[vn]): #, key = lambda x: abs(x), reverse = True):
                    print(''.join([ rx, cx, "%i<br /> %i" %( seg[0], seg[1]), xc, cx, format_end_time_in_utc(seg[0]), "<br />", format_end_time_in_utc(seg[1]), xc, cx, "%i" % abs(seg), xc, xr ]), file=summ_file)
                print(xt, file=summ_file) 
            print("%s\n%s" % ( '</body>', '</html>' ), file=summ_file)
            summ_file.close()

            # add to the correct file list
            if added_flags == [] and flag_intersects:
                undefined_flags[(this_ifo, this_name)] = summ_fn 
            elif added_flags == [] and flag_intersects:
                non_intersecting_flags[(this_ifo, this_name)] = summ_fn 
            else:
                updated_flags[(this_ifo, this_name)] = summ_fn


# write master summary file
all_ifos = sorted(set([v.ifo for v in veto_def_table]))
global_summ_fn = '%s/%s-VETO_DEFINER_CHECK_SUMMARY-%i-%i.html' %( opts.output_directory, ''.join(all_ifos), global_segment[0], abs(global_segment) )
summ_file = open( global_summ_fn, 'w' )
print("<html>\n<body>", file=summ_file)
print("<h1>Veto Definer Check: %i - %i </h1>" %(global_segment[0], global_segment[1]), file=summ_file)
print("<h2>(%s UTC - %s UTC)</h2>" %( format_end_time_in_utc(global_segment[0]), format_end_time_in_utc(global_segment[1]) ), file=summ_file)
print("<p>File(s) Checked:", file=summ_file)
print("<blockquote>", file=summ_file)
for filename, md5sum in analyzed_files.items():
    print('<a href="./%s">%s</a> md5sum: <code>%s</code>' % ( os.path.basename(filename), os.path.basename(filename), md5sum ), file=summ_file)
print("</blockquote>", file=summ_file)
if opts.write_new_veto_file:
    new_veto_fn = '%s-UPDATED_VETO_DEFINER-%i-%i.xml' %( ''.join(all_ifos), global_segment[0], abs(global_segment) )
    print('<p>New veto definer file written to: <blockquote><a href="./%s">%s</a></blockquote></p>' % ( new_veto_fn, new_veto_fn ), file=summ_file)
    print("<p>All veto entries corresponding to these flags were left un-altered:" + ulx, file=summ_file)
    print(lix + "All SciMon flags" + xli, file=summ_file)
    if opts.skip_undefined_flags:
        print(lix + "All Undefined flags" + xli, file=summ_file)
    for x in skip_ifos+skip_types:
        print(lix + "All %s flags" % x.upper() + xli, file=summ_file)
    for flag in skip_flags:
        if flag in check_names and flag not in other_skipped_flags:
            print(lix + ':'.join(flag) + xli, file=summ_file)
    print(xul + "</p>", file=summ_file)

print("<hr /><h2>SciMon Flags Checked</h2>", file=summ_file)
print('<p>These are flags that begin with "SCI-". "Found Flags Summary" links to the summary page with all vetoes in the veto definer file for which a corresponding flag could be found in the segment database during the desired times. All flags found in the segment database are listed on those pages. "Missing Flags Summary" links to the summary page with all vetoes for which a corresponding flag could <i>not</i> be found in the segment database during the desired times.</p>', file=summ_file)
if opts.write_new_veto_file:
    print('<p><b>All SciMon flags are written to the new veto definer file as they were originally, regardless of whether a corresponding flag could be found or not. To update these, please update by hand.</b></p>', file=summ_file)
print('<table border="1", cellpadding="5", align="center">', file=summ_file)
print(rx + thx + (xth+thx).join([ 'Flag', 'Found Flags Summary', 'Missing Flags Summary' ]), file=summ_file)
for (this_ifo, this_name) in sorted(scimon_flags):
    print(rx + cx + '%s:%s' % (this_ifo, this_name) + xc, end=' ', file=summ_file)
    if (this_ifo, this_name) in found_summ_files:
        print(''.join([ cx, hlx, "./summary_files/%s" % os.path.basename(found_summ_files[(this_ifo, this_name)]), hxl, 'link', xhl, xc]), end=' ', file=summ_file)
    else:
        print(''.join([ '<td style="background-color:#ff8080">', 'None', xc ]), end=' ', file=summ_file)
    if (this_ifo, this_name) in missing_summ_files:
        print(''.join([ '<td style="background-color:#ff8080">', hlx, "./summary_files/%s" % os.path.basename(missing_summ_files[(this_ifo, this_name)]), hxl, 'link', xhl, xc]), end=' ', file=summ_file)
    else:
        print(''.join([ '<td style="background-color:#80ff80">', 'None', xc ]), end=' ', file=summ_file)
    print(xr, file=summ_file)
print("</table>", file=summ_file)

print("<hr /><h2>Undefined Flags</h2>", file=summ_file)
print("<p>These are non-SciMon entries in the veto-definer file(s) that do not intersect with any defined segments at the correct version in the segment database, rendering them useless. This can happen for one of three reasons:<ol>", file=summ_file)
print(lix + "<b>Missing flags:</b> No summary segment could be found in the segment database during the times checked, implying that the flag is not in the segment database at all." + xli, file=summ_file)
print(lix + "<b>Non-intersecting flags:</b> The flag is defined in the database, but none of the defined segments intersect with times for which the veto was specified to be used. This could be because the version specified in the veto-definer file is higher than the maximum version found in the database; see the summary files for details." + xli, file=summ_file)
print(lix + "<b>Non-intersecting flags due to higher versions:</b> The entries in the veto-definer file do intersect with defined segments in the segment database at the versions specified but do not intersect with higher versions of the flag. This could indicate a mistake in the publication of the segments -- when the higher versions were added, segments occuring before the times that were meant to be fixed were not republished at the higher version, as they should have been -- or this could have been done on purpose -- the higher versions were added to the database because it was found that the flag should not be defined during those times, and so the veto should be removed. If the latter, leaving these flags in the veto-definer file will result in times being vetoed that shouldn't be. If the former, removing these flags will result in times not being vetoed that should be. <b>Without contacting the person who republished the segments, it is impossible to tell which is the case.</b>" + xli, file=summ_file)
print('</ol>', file=summ_file)

print("<h3>Missing Flags</h3>", file=summ_file)
if missing_flags == []:
    print('<p style="background-color:#80ff80; text-align:center"><br /><i>None</i><br />&nbsp</p>', file=summ_file)
else:
    if opts.write_new_veto_file and not opts.skip_undefined_flags:
        print("<p><b>All of these flags have been removed from the updated veto-definer file.</b></p>", file=summ_file)
    print('<table style="background-color:#ff8080" align="center", cellpadding="5">', file=summ_file)
    print(rx + thx + (xth+thx).join(["ifo", "name", "version", "category", "start_time", "end_time", "start_pad", "end_pad", "comment"]) + xth + xr, file=summ_file)
    for ifo, name in sorted(missing_flags):
        for veto in check_names[(ifo,name)]:
            print(rx + cx + (xc+cx).join([veto.ifo, veto.name, str(veto.version), str(veto.category), str(veto.start_time), str(veto.end_time), str(veto.start_pad), str(veto.end_pad), veto.comment]) + xc + xr, file=summ_file)
    print("</table>" + nl, file=summ_file)

print("<h3>Non-intersecting Flags</h3>", file=summ_file)
if list(non_intersecting_flags.keys()) == []:
    print('<p style="background-color:#80ff80; text-align:center"><br /><b>None</b><br />&nbsp</p>', file=summ_file)
else:
    if opts.write_new_veto_file:
        print("<p><b>All flags marked </b><i>removed</i><b> have been removed from the updated veto-definer file.</b></p>", file=summ_file)
    print('<table style="background-color:#ff8080"; align="center", cellpadding="5">', file=summ_file)
    print('<caption><i>Click on link to see summary</i></caption>', file=summ_file)
    updt_mark = '<td style="background-color:#80ff80"><i>removed</i>'
    for (this_ifo, this_name), summ_fn in sorted(non_intersecting_flags.items()):
        print(''.join([ rx, cx, hlx, "./summary_files/%s" % os.path.basename(summ_fn), hxl, "%s:%s" %( this_ifo, this_name), xhl, xc, opts.write_new_veto_file and '%s' %( ( opts.skip_undefined_flags or (this_ifo,this_name) in skip_flags) and cx or updt_mark) or cx, xc, xr ]), file=summ_file) 
    print(xt, file=summ_file)

print("<h3>Non-intersecting Flags due to Higher Versions</h3>", file=summ_file)
if list(undefined_flags.keys()) == []:
    print('<p style="background-color:#80ff80; text-align:center"><br /><b>None</b><br />&nbsp</p>', file=summ_file)
else:
    if opts.write_new_veto_file:
        print("<p><b>All flags marked </b><i>removed</i><b> have been removed from the updated veto-definer file.</b></p>", file=summ_file)
    print('<table style="background-color:#ff8080"; align="center", cellpadding="5">', file=summ_file)
    print('<caption><i>Click on link to see summary</i></caption>', file=summ_file)
    updt_mark = '<td style="background-color:#80ff80"><i>removed</i>'
    for (this_ifo, this_name), summ_fn in sorted(undefined_flags.items()):
        print(''.join([ rx, cx, hlx, "./summary_files/%s" % os.path.basename(summ_fn), hxl, "%s:%s" %( this_ifo, this_name), xhl, xc, opts.write_new_veto_file and '%s' %( ( opts.skip_undefined_flags or (this_ifo,this_name) in skip_flags) and cx or updt_mark) or cx, xc, xr ]), file=summ_file) 
    print(xt, file=summ_file)


print("<hr /><h2>Flags that Should be Updated</h2>", file=summ_file)
print("<p>These are non-SciMon flags that were found to have higher versions defined in the segment database.</p>", file=summ_file)
if list(undefined_flags.keys()) == []:
    print('<p style="background-color:#80ff80; text-align:center"><br /><b>None</b><br />&nbsp</p>', file=summ_file)
else:
    if opts.write_new_veto_file:
        print("<p><b>All flags marked </b><i>updated</i><b> have been updated in the new veto definer file according to what's shown on their respective summary pages.</b></p>", file=summ_file)
    print('<table style="background-color:#ff8080"; align="center", cellpadding="5">', file=summ_file)
    print('<caption><i>Click on link to see summary</i></caption>', file=summ_file)
    updt_mark = '<td style="background-color:#80ff80"><i>updated</i>'
    for (this_ifo, this_name), summ_fn in sorted(updated_flags.items()):
        print(''.join([ rx, cx, hlx, "./summary_files/%s" % os.path.basename(summ_fn), hxl, "%s:%s" %( this_ifo, this_name), xhl, xc, opts.write_new_veto_file and '%s' %( (this_ifo,this_name) in skip_flags and cx or updt_mark) or cx, xc, xr ]), file=summ_file)
    print(xt, file=summ_file)

print("<hr /><h2>Flags with No Higher Versions</h2>", file=summ_file)
print('<p>These are all non-SciMon flags for which there are no higher versions in the segment database than what is specified in the veto-definer file for all associated vetoes.</p><ul style="background-color:#80ff80"', file=summ_file)
for this_ifo, this_name in sorted(unchanged_flags):
    print(''.join([ lix, "%s:%s" % (this_ifo, this_name), xli ]), file=summ_file)
print(xul, file=summ_file)

print("<hr />", file=summ_file)
print(InspiralUtils.writeProcessParams( __prog__, git_version.verbose_msg.replace('\n','<br>'), sys.argv[1:] ).replace('Figure(s) produced', 'Results generated'), file=summ_file) 
print("%s\n%s" % ( '</body>', '</html>' ), file=summ_file)

summ_file.close()

# write xml file
if opts.write_new_veto_file:
    if opts.verbose:
        print("Writing new veto file...", file=sys.stderr)
    # add scimon flags
    for vl in scimon_flags.values():
        for veto in vl:
            outtable.append(veto)
    # add missing flags
    if not opts.skip_undefined_flags:
        outtable += [veto for this_ifo, this_name in check_names for veto in check_names[(this_ifo, this_name)] if (this_ifo, this_name) in missing_flags]
    outdoc.childNodes[0].appendChild(outtable)
    process.set_process_end_time(outproc_id)
    # try to find ligolw.xsl file based on the location of this executable
    if os.path.exists( '%s/%s' % (opts.output_directory, 'ligolw.xsl') ):
        xsl_file = 'ligolw.xsl'
    elif sys.argv[0].endswith('bin/%s' % __prog__) and os.path.exists(sys.argv[0].replace('bin/%s' % __prog__, 'etc/ligolw.xsl')):
        shutil.copy( sys.argv[0].replace('bin/%s' % __prog__, 'etc/ligolw.xsl'), opts.output_directory ) 
        xsl_file = 'ligolw.xsl'
    else:
        xsl_file = None
    utils.write_filename(outdoc, '%s/%s' %( opts.output_directory, new_veto_fn ), xsl_file = xsl_file)

if opts.verbose:
    print("Finished!\nSummary file written to:\n\t%s" % ( global_summ_fn ), file=sys.stderr)

sys.exit(0)



#
# xml dq publishing script for virgo (or other) dq xml files
#
#
# Copyright (C) 2009 Duncan Brown
# 
# This is part of the Grid LSC User Environment (GLUE)
# 
# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import pwd
import os
import re
import time
import logging
import logging.handlers

from optparse import OptionParser

from ligo import segments

from glue import gsiserverutils
from glue import gpstime

from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils as ligolw_utils
from glue.ligolw.utils import segments as ligolw_segments
from glue.ligolw.utils import process as ligolw_process
from glue.ligolw.utils import coalesce_db as ligolw_coalesce

from glue.segmentdb import segmentdb_utils


PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()

try:
  USER_NAME = os.getlogin()
except:
  USER_NAME = pwd.getpwuid(os.getuid())[0]

__author__  = "Ping Wei <piwei@physics.syr.edu>"
from glue import git_version
__date__    = git_version.date
__version__ = git_version.id

parser = OptionParser(
  version = "%prog CVS $Header$",
  usage   = "%prog [OPTIONS]",
  description = "Coalesce segments and summary intervals in the database")

parser.add_option("-d", "--database-name", metavar = "database_name", help = "name of the database you run coalescing on")
parser.add_option("-S", "--state-file", metavar = "FILE", help = "read coalesced time range from FILE")
parser.add_option("-P", "--pid-file", metavar = "FILE", help = "use FILE as process lock file")
parser.add_option("-l", "--log-file", metavar = "FILE", help = "use FILE as log file")
parser.add_option("-L", "--log-level", metavar = "LEVEL", default = "INFO", help = "set logging level to LEVEL")

options, filenames = parser.parse_args()

if not options.database_name:
  raise ValueError("missing argument --database-name")
if not options.state_file:
  raise ValueError("missing argument --state-file")
if not options.pid_file:
  raise ValueError("missing argument --pid-file")
if not options.log_file:
  raise ValueError("missing argument --log-file")

# check for an existing lock file
if os.access(options.pid_file,os.F_OK):
  raise RuntimeError("lock file exists: %s" % options.pid_file)

# create lock file
fp = open(options.pid_file, "w")
fp.write("lock")
fp.close()


try:
  # setup the output file
  outdoc = ligolw.Document()
  outdoc.appendChild(ligolw.LIGO_LW())
  proc_id = ligolw_process.register_to_xmldoc(outdoc, PROGRAM_NAME, options.__dict__).process_id

  # set up logging
  logger = logging.getLogger('segdb_coalesce')
  handler = logging.handlers.RotatingFileHandler(options.log_file, 'a', 1048576, 5)
  formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
  handler.setFormatter(formatter)
  logger.addHandler(handler)
  logger.setLevel(eval("logging." + options.log_level))
  logger.debug("pid file = " + options.pid_file)
  logger.debug("log file = " + options.log_file)
  logger.debug("log level = " + options.log_level)

  # read in the coalesced time range 
  logger.debug("reading state from %s" % options.state_file)
  indoc = ligolw_utils.load_url(options.state_file, gz = (options.state_file or "stdin").endswith(".gz"))
  coalesced_range = segmentdb_utils.find_segments(indoc,"P1:COALESCED:0")
  end_times = []
  for c in coalesced_range:
    end_times.append(c[1])
  max_endtime = max(end_times) 
  logger.debug("coalesced time range = %s" % str(coalesced_range))

  # make a list of the time range that need to be coalesced:
  #########################
  current_time = gpstime.GpsSecondsFromPyUTC(time.time())
  pending_range = [(max_endtime, current_time)]
  st = max_endtime
  et = current_time 
  logger.debug("pending range is [%d,%d)" % (st,et))

  # coalesce the pending range and add them to the list of coalesced range 
  logger.debug("start coalescing")
  result,data_existence = ligolw_coalesce.coalesce_seg(options.database_name,st,et)
  if result == 0 and data_existence == 1:
        coalesced_range |= segments.segmentlist([(st,et)])
        coalesced_id = segmentdb_utils.add_to_segment_definer(outdoc,proc_id,"P1","COALESCED",0)
        segmentdb_utils.add_to_segment(outdoc,proc_id,coalesced_id,coalesced_range)

        # write the new segment state file on top of the old one
        ligolw_utils.write_filename(outdoc, options.state_file)

        logger.info("finished coalescing range [%d,%d)" % (st,et))
        logger.info("exiting")
  elif result == 0 and data_existence != 1:
        logger.info("no data found in time range, state_file remians unchanged")
  if result != 0 :
        logger.error("fail to coalesce, state_file remains unchanged")
        try:
          os.unlink(options.pid_file)
        except:
          pass
        logger.error(result)
        sys.exit(1)

except Exception as e:
  try:
    logger.error(str(e))
  except:
    pass
  print("runtime error (%s)" % str(e), file=sys.stderr)
  os.unlink(options.pid_file)
  sys.exit(1)

os.unlink(options.pid_file)
sys.exit(0)




#
# xml dq publishing script for virgo (or other) dq xml files
#
#
# Copyright (C) 2009 Duncan Brown
# 
# This is part of the Grid LSC User Environment (GLUE)
# 
# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import pwd
import os
import re
import signal
import re
import time
import glob
import logging
import logging.handlers

from optparse import OptionParser

from ligo import segments
from glue import lal
from glue import gpstime
from ligo.segments import utils as segmentsUtils
from glue import pidfile

from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils as ligolw_utils
from glue.ligolw.utils import segments as ligolw_segments
from glue.ligolw.utils import process as ligolw_process

from glue.segmentdb import segmentdb_utils

# try and exit gracefully on a term signal
die_now = False

def SIGTERMhandler(signum, frame):
  global die_now
  try:
    logger.info("caught SIGTERM")
  except:
    pass
  die_now = True

signal.signal(signal.SIGTERM, SIGTERMhandler)

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()

try:
  USER_NAME = os.getlogin()
except:
  USER_NAME = pwd.getpwuid(os.getuid())[0]

from glue import git_version
__author__  = "Duncan Brown <dabrown@physics.syr.edu>"
__date__    = git_version.date
__version__ = git_version.id

parser = OptionParser(
  version = "%prog CVS $Header$",
  usage   = "%prog [OPTIONS]",
  description = "Publishes XML files into the segment database")

parser.add_option("-t", "--segment-url", metavar = "PROTOCOL://HOST", help = "connect to ldbd on PROTOCOL://HOST")
parser.add_option("-s", "--state-file", metavar = "FILE", help = "read published and excluded segments from FILE")
parser.add_option("-f", "--segments-file", metavar = "FILE", help = "read list of segments to publish from FILE")
parser.add_option("-P", "--pid-file", metavar = "FILE", help = "use FILE as process lock file")
parser.add_option("-D", "--input-directory", metavar = "DIR", help = "look for input files in DIR")
parser.add_option("-l", "--log-file", metavar = "FILE", help = "use FILE as log file")
parser.add_option("-L", "--log-level", metavar = "LEVEL", default = "INFO", help = "set logging level to LEVEL")
parser.add_option("-p", "--ping", action = "store_true")
parser.add_option("-d", "--dry-run", action = "store_true")

options, filenames = parser.parse_args()

if not options.segment_url:
  raise ValueError("missing argument --segment-url")

# open connection to LDBDWClient
myClient = segmentdb_utils.setup_database(options.segment_url)


if options.ping:
  msg = myClient.ping()
  print(msg)
  sys.exit(0)

if not options.state_file and not options.segments_file:
  raise ValueError("missing argument --state-file or --segments-file")
if options.state_file and options.segments_file:
  raise ValueError("incompatible arguments --state-file and --segments-file")
if not options.pid_file:
  raise ValueError("missing argument --pid-file")
if not options.input_directory:
  raise ValueError("missing argument --input-directory")
if not options.log_file:
  raise ValueError("missing argument --log-file")

# check if a valid lock file already exists, and create one if not
pidfile.get_lock(options.pid_file)

try:
  # set up logging
  logger = logging.getLogger('ligolw_publish_dqxml')
  handler = logging.handlers.RotatingFileHandler(options.log_file, 'a', 1024**3, 3)
  formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
  handler.setFormatter(formatter)
  logger.addHandler(handler)
  logger.setLevel(eval("logging." + options.log_level))
  logger.info("ligolw_publish_dqxml starting")
  logger.debug("pid file = " + options.pid_file)
  logger.debug("input directory = " + options.input_directory)
  logger.debug("log file = " + options.log_file)
  logger.debug("log level = " + options.log_level)

  if options.state_file:
    logger.debug("segment-url= " + options.segment_url)
    # setup the output file
    outdoc = ligolw.Document()
    outdoc.appendChild(ligolw.LIGO_LW())
    proc_id = ligolw_process.register_to_xmldoc(outdoc, PROGRAM_NAME, options.__dict__).process_id

    # read in the published file and get the published segment list
    logger.debug("reading state from %s" % options.state_file)
    indoc = ligolw_utils.load_url(options.state_file, gz = (options.state_file or "stdin").endswith(".gz"))
    published_segments = segmentdb_utils.find_segments(indoc,"P1:PUBLISHED:0")
    exclude_segments = segmentdb_utils.find_segments(indoc,"P1:EXCLUDE:0")
    logger.debug("published segments = %s" % str(published_segments))
    logger.debug("excluded segments = %s" % str(exclude_segments))
    # FIXME this will break Jul 19 2027 02:39:45 UTC
    all_time = segments.segmentlist([segments.segment(0,1500000000)])
    # make a list of the segments that need to be inserted
    pending_segments = (all_time - published_segments) - exclude_segments

  elif options.segments_file:
    logger.debug("segments-file= " + options.segments_file)
    logger.debug("reading segments from %s" % options.segments_file)
    # read in the segments file and get the list of segments to insert
    sfile = open(options.segments_file, "r")
    pending_segments = segmentsUtils.fromsegwizard(sfile).coalesce()
    sfile.close()
    published_segments = segments.segmentlist()

  logger.info("pending segments = %s" % str(pending_segments))
  pending_files = lal.Cache()
  # make a list of the files that need to be inserted
  for s in pending_segments:
    pending_files += lal.Cache.from_urls(segmentdb_utils.get_all_files_in_range(options.input_directory,s[0],s[1]),coltype=int).sieve(segment=s)
  pending_files = pending_files.unique()
  pending_files.sort()

  logger.debug("pending files = %s" % [os.path.basename(f) for f in pending_files.pfnlist()])

  # publish the files and add them to the list of published segments
  for f in pending_files:
    result = None
    if die_now:
      break
    infile = f.path
    try:
      logger.debug("reading %s" % infile)
      fh = open(infile,'r')
      xmltext = fh.read()
      fh.close()
      if options.dry_run:
        logger.debug("(dry-run) inserting %s" % infile)
        result = True
      else:
        logger.debug("inserting %s" % infile)
        myClient = segmentdb_utils.setup_database(options.segment_url)
        result = myClient.insertdmt(xmltext)
      del xmltext
    except KeyboardInterrupt:
      logger.info("caught keyboard interrupt")
      if result:
        logger.debug("published segment %s" % str(f.segment))
        published_segments |= segments.segmentlist([f.segment])
      die_now = True
    except Exception as e:
      logger.error("failed to publish %s (%s)" % (infile, str(e)))
    else:
      logger.debug("published segment %s" % str(f.segment))
      published_segments |= segments.segmentlist([f.segment])

  if options.state_file:
    excl_def_id = segmentdb_utils.add_to_segment_definer(outdoc,proc_id,"P1","EXCLUDE",0)
    pub_def_id = segmentdb_utils.add_to_segment_definer(outdoc,proc_id,"P1","PUBLISHED",0)
    segmentdb_utils.add_to_segment(outdoc,proc_id,excl_def_id,exclude_segments)
    segmentdb_utils.add_to_segment(outdoc,proc_id,pub_def_id,published_segments)
    logger.debug("published segments = %s" % str(published_segments))
    logger.debug("excluded segments = %s" % str(exclude_segments))

    # write the new segment state file on top of the old one
    logger.debug("writing state to %s" % options.state_file)
    ligolw_utils.write_filename(outdoc, options.state_file)

except Exception as e:
  try:
    logger.error(str(e))
  except:
    pass
  print("runtime error (%s)" % str(e), file=sys.stderr)
  os.unlink(options.pid_file)
  sys.exit(1)

logger.info("exiting")
os.unlink(options.pid_file)
sys.exit(0)




#
# lightweight database dumper client
#
#
# Copyright (C) 2003 Duncan Brown
# 
# This program is part of the Grid LSC User Environment (GLUE)
# 
# GLUE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
# 
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys
import os
import getopt
import re

try:
  from glue.segmentdb import segmentdb_utils
except ImportError as e:
  print("""
Error: unable to import modules from glue.

Check that glue is correctly installed and in your PYTHONPATH.

%s
""" % e, file=sys.stderr)
  sys.exit(1)


def usage():
  msg = """\
Usage: ldbdc.py [OPTIONS]
  -t, --segment-url       Users have to specify either 'https://' for a secure connection or 'http://' for an insecure connection in the segment database url. For example, '--segment-url=https://segdb.ligo.caltech.edu'. No need to specify port number 
  -d, --database          use database specified by environment variable S6_SEGMENT_SERVER. For example, 'S6_SEGMENT_SERVER=https://segdb.ligo.caltech.edu'"
  -i, --insert LIGOLW     insert LIGO lightweight file LIGOLW
  -f, --pfn PFNLIST       a comma separated list of PFNs to insert
  -q, --query QUERY       execute SQL query QUERY
  -o, --output FILE       write output to FILE
  -D, --dmt-segments      insert the XML data as DMT segment data
  -p, --ping              ping the server
  -h, --help              print this message

One of the options --ping, --query, --insert must be specified.
\
"""
  print(msg)

def help():
  usage()
  sys.exit(0)

shortop = "t:i:f:q:o:dDph"
longop = [
  "segment-url=",
  "identity=",
  "database",
  "insert=",
  "pfn=",
  "query=",
  "output=",
  "dmt-segments",
  "ping",
  "help"
  ]

try:
  opts, args = getopt.getopt(sys.argv[1:], shortop, longop)
except getopt.GetoptError:
  print("Error: %s \n" % str(e), file=sys.stderr)
  usage()
  sys.exit(1)

# defaults
segment_url = None
infile = None
pfnlist = None
outfile = None
squery = None
database = None
dmtsegs = None
ping = None

# environment variables override defaults but not command line options
# This block of code is also required for web pages calling ldbdc to work
try:
  segment_url = os.environ['LDBD_SERVER']
except:
  pass

for o, a in opts:
  if o in ("-h", "--help"):
    help()
  elif o in ("-t", "--segment-url"):
    segment_url = a
  elif o in ("-i", "--insert"):
    infile = a
  elif o in ("-f", "--pfn"):
    pfnlist = a
  elif o in ("-q", "--query"):
    squery = a
  elif o in ("-d", "--database"):
    database = os.environ['S6_SEGMENT_SERVER']
  elif o in ("-o", "--output"):
    outfile = a
  elif o in ("-D", "--dmt-segments"):
    dmtsegs = 1
  elif o in ("-p", "--ping"):
    ping = 1


if segment_url and database:
  print("\nMust specify either --segment-url or --database")
  usage()
  sys.exit(1)
if not segment_url and not database:
  print("\nMust specify either --segment-url or --database")
  usage()
  sys.exit(1)
if not infile and not squery and not ping:
  print("\nNo operation specified\n")
  usage()
  sys.exit(1)
if squery and infile:
  print("\nOnly one of --query and --insert can be specified\n")
  usage()
  sys.exit(1)
if pfnlist and not infile:
  print("\nA list of PFNs makes sense only when inserting data\n")
  usage()
  sys.exit(1)
if dmtsegs and not infile:
  print("\nMust specify an XML file to insert\n")
  usage()
  sys.exit(1)
if dmtsegs and pfnlist:
  print("\nCannot specify a list of PFNs when inserting data segment data\n")
  usage()
  sys.exit(1)


# open connection to LDBD(W)Server
myClient = segmentdb_utils.setup_database(segment_url)


try:
  if ping:
    # ping the server and print the response
    print(myClient.ping())

  elif infile and not pfnlist:
    # open an input file and send it to the server
    fh = open(infile,'r')
    xmltext = fh.read()
    if dmtsegs:
      myClient.insertdmt(xmltext)
    else:
      myClient.insert(xmltext)
    fh.close()

  elif infile and pfnlist:
    # open the input file but skip any of the "bad" tables
    xmltext = ""
    p = LDBDClient.SimpleLWXMLParser()
    fh = open(infile,'r')
    for line in fh:
      xmltext += p.parse_line(line)
    fh.close()
    lfnpfn_dict = {}
    lfnpfn_dict[os.path.basename(infile)] = pfnlist.split(',')
    myClient.insertmap(xmltext,lfnpfn_dict)

  elif squery:
    # execute the query and write out the results
    if outfile:
      fh = open(outfile,'w')
    else:
      fh = sys.__stdout__
    fh.write(myClient.query(squery))
    fh.close()

except Exception as e:
  print("Error querying LDBD Server: %s" % str(e), file=sys.stderr)
  print("Enter 'ldbdc --help' for usage", file=sys.stderr)
  sys.exit(1)

sys.exit(0)
